import pytest

import numpy as np

import nengo
from nengo.exceptions import ValidationError
from nengo.processes import WhiteSignal
from nengo.utils.numpy import rms

from nengolib.networks.rolling_window import (
    _pade_readout, _legendre_readout, t_default, RollingWindow)
from nengolib import Network
from nengolib.signal import (LinearSystem, shift, nrmse, Identity,
                             EvalPoints, Encoders)
from nengolib.stats import cube
from nengolib.synapses import PadeDelay, LegendreDelay


@pytest.mark.parametrize("legendre", [True, False])
@pytest.mark.parametrize("d,tol", [
    (2, 0.5), (3, 0.2), (4, 0.05), (5, 1e-2), (7, 1e-4), (12, 1e-9)])
def test_readout(legendre, d, tol):
    theta = 0.1

    if legendre:
        sys = LegendreDelay(theta, d)
        readout = _legendre_readout
    else:
        sys = PadeDelay(theta, d)
        readout = _pade_readout
    # decoding at r=1 (t=-theta) is equivalent to decoding a delay of theta
    C = readout(d, 1)
    assert np.allclose(sys.C, C)
    assert np.allclose(sys.D, 0)

    freqs = np.linspace(0, 5, 100)
    s = 2j * np.pi * freqs

    # check that frequency response has small error at low-frequencies
    # for a variety of different readouts
    for r in np.linspace(0, 1, 100):
        C = readout(d, r)

        sys = LinearSystem((sys.A, sys.B, C, sys.D), analog=True)
        error = np.abs(sys(s) - np.exp(-r*theta*s))
        assert 0 < rms(error) < tol, r


@pytest.mark.parametrize("legendre", [True, False])
def test_direct_window(legendre, Simulator, seed, plt):
    theta = 1.0
    T = theta

    assert np.allclose(t_default, np.linspace(0, 1, 1000))

    with Network() as model:
        stim = nengo.Node(output=lambda t: t)
        rw = RollingWindow(theta, n_neurons=1, dimensions=12,
                           neuron_type=nengo.Direct(), process=None,
                           legendre=legendre)
        assert rw.theta == theta
        assert rw.dt == 0.001
        assert rw.process is None
        assert rw.synapse == nengo.Lowpass(0.1)
        assert rw.input_synapse == nengo.Lowpass(0.1)

        nengo.Connection(stim, rw.input, synapse=None)
        output = rw.add_output(function=lambda w: np.sum(w**3)**2)
        p_output = nengo.Probe(output, synapse=None)

    with Simulator(model) as sim:
        sim.run(T)

    actual = sim.data[p_output].squeeze()
    t = sim.trange()
    ideal = shift(np.cumsum(t**3)**2)

    plt.figure()
    plt.plot(t, actual, label="Output")
    plt.plot(t, ideal, label="Ideal", linestyle='--')
    plt.legend()

    assert nrmse(actual, ideal) < 0.005


@pytest.mark.parametrize("legendre", [True, False])
def test_basis(legendre, plt):
    theta = 0.1
    d = 12

    rw = RollingWindow(theta=theta, n_neurons=1, dimensions=d, radii=2,
                       realizer=Identity(), legendre=legendre, process=None)

    B = rw.basis()
    assert B.shape == (len(t_default), d)

    # since using the Identity realizer
    assert np.allclose(B, rw.canonical_basis()*rw.radii)

    Binv = rw.inverse_basis()
    assert Binv.shape == (d, len(t_default))

    assert np.allclose(Binv.dot(B), np.eye(d))

    x = rw.sys.X.impulse(100)
    assert np.allclose(Binv.dot(B.dot(x.T)), x.T)

    plt.subplot(2, 1, 1)
    plt.plot(B)
    plt.subplot(2, 1, 2)
    plt.plot(Binv.T)


@pytest.mark.parametrize("legendre", [True, False])
def test_window_example(legendre, Simulator, seed, plt):
    theta = 0.1
    n_neurons = 1000
    d = 6
    high = 10

    T = 1.0
    dt = 0.002
    tau_probe = None

    # we set the radii here for testing on nengo<2.4.0
    # (see warning in RollingWindow._make_core)
    radii = 0.3

    with Network(seed=seed) as model:
        stim = nengo.Node(output=WhiteSignal(T, high=high, seed=seed))

        rw_rate = RollingWindow(
            theta=theta, n_neurons=n_neurons, dimensions=d, radii=radii,
            neuron_type=nengo.LIFRate(), dt=dt, process=stim.output,
            legendre=legendre)
        assert isinstance(rw_rate.state.eval_points, EvalPoints)
        assert isinstance(rw_rate.state.encoders, Encoders)

        rw_drct = RollingWindow(
            theta=theta, n_neurons=1, dimensions=d, radii=radii,
            neuron_type=nengo.Direct(), dt=dt, process=None,
            legendre=legendre)

        def function(w):
            return -np.max(np.abs(w)), abs(w[0]*w[-1])

        nengo.Connection(stim, rw_rate.input, synapse=None)
        nengo.Connection(stim, rw_drct.input, synapse=None)

        delay_rate = rw_rate.add_output(t=1)
        delay_drct = rw_drct.output  # rw_drct.add_output('delay', t=1)

        output_rate = rw_rate.add_output(function=function)
        output_drct = rw_drct.add_output(function=function)

        p_stim = nengo.Probe(stim, synapse=tau_probe)
        p_delay_rate = nengo.Probe(delay_rate, synapse=tau_probe)
        p_delay_drct = nengo.Probe(delay_drct, synapse=tau_probe)
        p_output_rate = nengo.Probe(output_rate, synapse=tau_probe)
        p_output_drct = nengo.Probe(output_drct, synapse=tau_probe)
        p_state_rate = nengo.Probe(rw_rate.state, synapse=tau_probe)
        p_state_drct = nengo.Probe(rw_drct.state, synapse=tau_probe)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    plt.subplot(3, 1, 1)
    plt.plot(sim.trange(), sim.data[p_stim], label="Input")
    plt.plot(sim.trange(), sim.data[p_delay_rate], label="Rate")
    plt.plot(sim.trange(), sim.data[p_delay_drct], label="Direct",
             linestyle='--')
    plt.legend()

    plt.subplot(3, 1, 2)
    plt.plot(sim.trange(), sim.data[p_output_rate], label="Rate")
    plt.plot(sim.trange(), sim.data[p_output_drct], label="Direct",
             linestyle='--')
    plt.legend()

    plt.subplot(3, 1, 3)
    plt.plot(sim.trange(), sim.data[p_state_rate], label="Rate", lw=1)
    plt.plot(sim.trange(), sim.data[p_state_drct], label="Direct",
             linestyle='--')

    assert nrmse(sim.data[p_delay_rate], sim.data[p_delay_drct]) < 0.05
    assert nrmse(sim.data[p_output_rate], sim.data[p_output_drct]) < 0.2
    assert nrmse(sim.data[p_state_rate], sim.data[p_state_drct]) < 0.05


def test_window_function():
    process = WhiteSignal(1.0, 10)
    with pytest.raises(ValidationError):
        RollingWindow(theta=1.0, n_neurons=1, process=process,
                      eval_points=cube)

    rw = RollingWindow(theta=1.0, n_neurons=1, process=None)
    with pytest.raises(ValidationError):
        rw.add_output(function=lambda a, b: a)



import numpy as np
import pytest

import nengo
from nengo.builder import Builder
from nengo.builder.operator import Reset, Copy
from nengo.builder.signal import Signal
from nengo.dists import UniformHypersphere
from nengo.exceptions import ValidationError
from nengo.learning_rules import LearningRuleTypeParam, PES, BCM, Oja, Voja
from nengo.processes import WhiteSignal
from nengo.synapses import Alpha, Lowpass


def best_weights(weight_data):
    return np.argmax(np.sum(np.var(weight_data, axis=0), axis=0))


def _test_pes(Simulator, nl, plt, seed,
              pre_neurons=False, post_neurons=False, weight_solver=False,
              vin=np.array([0.5, -0.5]), vout=None, n=200,
              function=None, transform=np.array(1.), rate=1e-3):

    vout = np.array(vin) if vout is None else vout

    model = nengo.Network(seed=seed)
    with model:
        model.config[nengo.Ensemble].neuron_type = nl()

        u = nengo.Node(output=vin)
        v = nengo.Node(output=vout)
        a = nengo.Ensemble(n, dimensions=u.size_out)
        b = nengo.Ensemble(n, dimensions=u.size_out)
        e = nengo.Ensemble(n, dimensions=v.size_out)

        nengo.Connection(u, a)

        bslice = b[:v.size_out] if v.size_out < u.size_out else b
        pre = a.neurons if pre_neurons else a
        post = b.neurons if post_neurons else bslice

        conn = nengo.Connection(pre, post,
                                function=function, transform=transform,
                                learning_rule_type=PES(rate))
        if weight_solver:
            conn.solver = nengo.solvers.LstsqL2(weights=True)

        nengo.Connection(v, e, transform=-1)
        nengo.Connection(bslice, e)
        nengo.Connection(e, conn.learning_rule)

        b_p = nengo.Probe(bslice, synapse=0.03)
        e_p = nengo.Probe(e, synapse=0.03)

        weights_p = nengo.Probe(conn, 'weights', sample_every=0.01)
        corr_p = nengo.Probe(conn.learning_rule, 'correction', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)
    t = sim.trange()
    weights = sim.data[weights_p]

    plt.subplot(311)
    plt.plot(t, sim.data[b_p])
    plt.ylabel("Post decoded value")
    plt.subplot(312)
    plt.plot(t, sim.data[e_p])
    plt.ylabel("Error decoded value")
    plt.subplot(313)
    plt.plot(t, sim.data[corr_p] / rate)
    plt.ylabel("PES correction")
    plt.xlabel("Time (s)")

    tend = t > 0.4
    assert np.allclose(sim.data[b_p][tend], vout, atol=0.05)
    assert np.allclose(sim.data[e_p][tend], 0, atol=0.05)
    assert np.allclose(sim.data[corr_p][tend] / rate, 0, atol=0.05)
    assert not np.allclose(weights[0], weights[-1], atol=1e-5)


def test_pes_ens_ens(Simulator, nl_nodirect, plt, seed):
    function = lambda x: [x[1], x[0]]
    _test_pes(Simulator, nl_nodirect, plt, seed, function=function)


def test_pes_weight_solver(Simulator, plt, seed):
    function = lambda x: [x[1], x[0]]
    _test_pes(Simulator, nengo.LIF, plt, seed, function=function,
              weight_solver=True)


def test_pes_ens_slice(Simulator, plt, seed):
    vin = [0.5, -0.5]
    vout = [vin[0]**2 + vin[1]**2]
    function = lambda x: [x[0] - x[1]]
    _test_pes(Simulator, nengo.LIF, plt, seed, vin=vin, vout=vout,
              function=function)


def test_pes_neuron_neuron(Simulator, plt, seed, rng):
    n = 200
    initial_weights = rng.uniform(high=4e-4, size=(n, n))
    _test_pes(Simulator, nengo.LIF, plt, seed,
              pre_neurons=True, post_neurons=True,
              n=n, transform=initial_weights)


def test_pes_neuron_ens(Simulator, plt, seed, rng):
    n = 200
    initial_weights = rng.uniform(high=1e-4, size=(2, n))
    _test_pes(Simulator, nengo.LIF, plt, seed,
              pre_neurons=True, post_neurons=False,
              n=n, transform=initial_weights)


def test_pes_transform(Simulator, seed):
    """Test behaviour of PES when function and transform both defined."""
    n = 200
    # error must be with respect to transformed vector (conn.size_out)
    T = np.asarray([[0.5], [-0.5]])  # transform to output

    m = nengo.Network(seed=seed)
    with m:
        u = nengo.Node(output=[1])
        a = nengo.Ensemble(n, dimensions=1)
        b = nengo.Node(size_in=2)
        e = nengo.Node(size_in=1)

        nengo.Connection(u, a)
        learned_conn = nengo.Connection(
            a, b, function=lambda x: [0], transform=T,
            learning_rule_type=nengo.PES(learning_rate=1e-3))
        assert T.shape[0] == learned_conn.size_out
        assert T.shape[1] == learned_conn.size_mid

        nengo.Connection(b[0], e, synapse=None)
        nengo.Connection(nengo.Node(output=-1), e)
        nengo.Connection(
            e, learned_conn.learning_rule, transform=T, synapse=None)

        p_b = nengo.Probe(b, synapse=0.05)

    with Simulator(m) as sim:
        sim.run(1.0)
    tend = sim.trange() > 0.7

    assert np.allclose(sim.data[p_b][tend], [1, -1], atol=1e-2)


def test_pes_multidim_error(Simulator, seed):
    """Test that PES works on error connections mapping from N to 1 dims.

    Note that the transform is applied before the learning rule, so the error
    signal should be 1-dimensional.
    """

    with nengo.Network(seed=seed) as net:
        err = nengo.Node(output=[0])
        ens1 = nengo.Ensemble(20, 3)
        ens2 = nengo.Ensemble(10, 1)

        # Case 1: ens -> ens, weights=False
        conn = nengo.Connection(ens1, ens2,
                                transform=np.ones((1, 3)),
                                solver=nengo.solvers.LstsqL2(weights=False),
                                learning_rule_type={"pes": nengo.PES()})
        nengo.Connection(err, conn.learning_rule["pes"])
        # Case 2: ens -> ens, weights=True
        conn = nengo.Connection(ens1, ens2,
                                transform=np.ones((1, 3)),
                                solver=nengo.solvers.LstsqL2(weights=True),
                                learning_rule_type={"pes": nengo.PES()})
        nengo.Connection(err, conn.learning_rule["pes"])
        # Case 3: neurons -> ens
        conn = nengo.Connection(ens1.neurons, ens2,
                                transform=np.ones((1, ens1.n_neurons)),
                                learning_rule_type={"pes": nengo.PES()})
        nengo.Connection(err, conn.learning_rule["pes"])

    with Simulator(net) as sim:
        sim.run(0.01)


@pytest.mark.parametrize('pre_synapse', [
    0, Lowpass(tau=0.05), Alpha(tau=0.005)])
def test_pes_synapse(Simulator, seed, pre_synapse):
    rule = PES(pre_synapse=pre_synapse)

    with nengo.Network(seed=seed) as model:
        stim = nengo.Node(output=WhiteSignal(0.5, high=10))
        x = nengo.Ensemble(100, 1)

        nengo.Connection(stim, x, synapse=None)
        conn = nengo.Connection(x, x, learning_rule_type=rule)

        p_neurons = nengo.Probe(x.neurons, synapse=pre_synapse)
        p_pes = nengo.Probe(conn.learning_rule, 'activities')

    with Simulator(model) as sim:
        sim.run(0.5)

    assert np.allclose(sim.data[p_neurons][1:, :],
                       sim.data[p_pes][:-1, :])


@pytest.mark.parametrize("weights", [False, True])
def test_pes_recurrent_slice(Simulator, seed, weights):
    """Test that PES works on recurrent connections from N to 1 dims."""

    with nengo.Network(seed=seed) as net:
        err = nengo.Node(output=[-10])
        stim = nengo.Node(output=[0, 0])
        post = nengo.Ensemble(20, 2)
        nengo.Connection(stim, post)

        conn = nengo.Connection(post, post[1],
                                function=lambda x: 0.0,
                                solver=nengo.solvers.LstsqL2(weights=weights),
                                learning_rule_type=nengo.PES())

        nengo.Connection(err, conn.learning_rule)
        p = nengo.Probe(post, synapse=0.025)

    with Simulator(net) as sim:
        sim.run(0.1)

    # Learning rule should drive second dimension high, but not first
    assert np.all(sim.data[p][-10:, 0] < 0.2)
    assert np.all(sim.data[p][-10:, 1] > 0.8)


@pytest.mark.parametrize('rule_type, solver', [
    (BCM(learning_rate=1e-8), False),
    (Oja(learning_rate=1e-5), False),
    ([Oja(learning_rate=1e-5), BCM(learning_rate=1e-8)], False),
    ([Oja(learning_rate=1e-5), BCM(learning_rate=1e-8)], True),
])
def test_unsupervised(Simulator, rule_type, solver, seed, rng, plt):
    n = 200

    m = nengo.Network(seed=seed)
    with m:
        u = nengo.Node(WhiteSignal(0.5, high=10), size_out=2)
        a = nengo.Ensemble(n, dimensions=2)
        b = nengo.Ensemble(n+1, dimensions=2)
        nengo.Connection(u, a)

        if solver:
            conn = nengo.Connection(
                a, b, solver=nengo.solvers.LstsqL2(weights=True))
        else:
            initial_weights = rng.uniform(
                high=1e-3, size=(b.n_neurons, a.n_neurons))
            conn = nengo.Connection(
                a.neurons, b.neurons, transform=initial_weights)
        conn.learning_rule_type = rule_type

        inp_p = nengo.Probe(u)
        weights_p = nengo.Probe(conn, 'weights', sample_every=0.01)

        ap = nengo.Probe(a, synapse=0.03)
        up = nengo.Probe(b, synapse=0.03)

    with Simulator(m, seed=seed+1) as sim:
        sim.run(0.5)
    t = sim.trange()

    plt.subplot(2, 1, 1)
    plt.plot(t, sim.data[inp_p], label="Input")
    plt.plot(t, sim.data[ap], label="Pre")
    plt.plot(t, sim.data[up], label="Post")
    plt.legend(loc="best", fontsize="x-small")
    plt.subplot(2, 1, 2)
    best_ix = best_weights(sim.data[weights_p])
    plt.plot(sim.trange(dt=0.01), sim.data[weights_p][..., best_ix])
    plt.xlabel("Time (s)")
    plt.ylabel("Weights")

    assert not np.allclose(sim.data[weights_p][0], sim.data[weights_p][-1])


def learning_net(learning_rule=nengo.PES, net=None, rng=np.random):
    net = nengo.Network() if net is None else net
    with net:
        if learning_rule is nengo.PES:
            learning_rule_type = learning_rule(learning_rate=1e-5)
        else:
            learning_rule_type = learning_rule()

        u = nengo.Node(output=1.0)
        pre = nengo.Ensemble(10, dimensions=1)
        post = nengo.Ensemble(10, dimensions=1)
        initial_weights = rng.uniform(high=1e-3,
                                      size=(pre.n_neurons, post.n_neurons))
        conn = nengo.Connection(pre.neurons, post.neurons,
                                transform=initial_weights,
                                learning_rule_type=learning_rule_type)
        if learning_rule is nengo.PES:
            err = nengo.Ensemble(10, dimensions=1)
            nengo.Connection(u, err)
            nengo.Connection(err, conn.learning_rule)

        net.activity_p = nengo.Probe(pre.neurons, synapse=0.01)
        net.weights_p = nengo.Probe(
            conn, 'weights', synapse=None, sample_every=.01)
    return net


@pytest.mark.parametrize('learning_rule', [nengo.PES, nengo.BCM, nengo.Oja])
def test_dt_dependence(Simulator, plt, learning_rule, seed, rng):
    """Learning rules should work the same regardless of dt."""
    m = learning_net(learning_rule, nengo.Network(seed=seed), rng)

    trans_data = []
    # Using dts greater near tau_ref (0.002 by default) causes learning to
    # differ due to lowered presynaptic firing rate
    dts = (0.0001, 0.001)
    colors = ('b', 'g', 'r')
    ax1 = plt.subplot(2, 1, 1)
    ax2 = plt.subplot(2, 1, 2)
    for c, dt in zip(colors, dts):
        with Simulator(m, dt=dt) as sim:
            sim.run(0.1)
        trans_data.append(sim.data[m.weights_p])
        best_ix = best_weights(sim.data[m.weights_p])
        ax1.plot(sim.trange(dt=0.01), sim.data[m.weights_p][..., best_ix], c=c)
        ax2.plot(sim.trange(), sim.data[m.activity_p], c=c)

    ax1.set_xlim(right=sim.trange()[-1])
    ax1.set_ylabel("Connection weight")
    ax2.set_xlim(right=sim.trange()[-1])
    ax2.set_ylabel("Presynaptic activity")

    assert np.allclose(trans_data[0], trans_data[1], atol=3e-3)
    assert not np.allclose(sim.data[m.weights_p][0], sim.data[m.weights_p][-1])


@pytest.mark.parametrize('learning_rule', [nengo.PES, nengo.BCM, nengo.Oja])
def test_reset(Simulator, learning_rule, plt, seed, rng):
    """Make sure resetting learning rules resets all state."""
    m = learning_net(learning_rule, nengo.Network(seed=seed), rng)

    with Simulator(m) as sim:
        sim.run(0.1)
        sim.run(0.2)

        first_t = sim.trange()
        first_t_trans = sim.trange(dt=0.01)
        first_activity_p = np.array(sim.data[m.activity_p], copy=True)
        first_weights_p = np.array(sim.data[m.weights_p], copy=True)

        sim.reset()
        sim.run(0.3)

    plt.subplot(2, 1, 1)
    plt.ylabel("Neural activity")
    plt.plot(first_t, first_activity_p, c='b')
    plt.plot(sim.trange(), sim.data[m.activity_p], c='g')
    plt.subplot(2, 1, 2)
    plt.ylabel("Connection weight")
    best_ix = best_weights(first_weights_p)
    plt.plot(first_t_trans, first_weights_p[..., best_ix], c='b')
    plt.plot(sim.trange(dt=0.01), sim.data[m.weights_p][..., best_ix], c='g')

    assert np.allclose(sim.trange(), first_t)
    assert np.allclose(sim.trange(dt=0.01), first_t_trans)
    assert np.allclose(sim.data[m.activity_p], first_activity_p)
    assert np.allclose(sim.data[m.weights_p], first_weights_p)


def test_learningruletypeparam():
    """LearningRuleTypeParam must be one or many learning rules."""
    class Test(object):
        lrp = LearningRuleTypeParam('lrp', default=None)

    inst = Test()
    assert inst.lrp is None
    inst.lrp = Oja()
    assert isinstance(inst.lrp, Oja)
    inst.lrp = [Oja(), Oja()]
    for lr in inst.lrp:
        assert isinstance(lr, Oja)
    # Non-LR no good
    with pytest.raises(ValueError):
        inst.lrp = 'a'
    # All elements in list must be LR
    with pytest.raises(ValueError):
        inst.lrp = [Oja(), 'a', Oja()]


def test_learningrule_attr(seed):
    """Test learning_rule attribute on Connection"""
    def check_rule(rule, conn, rule_type):
        assert rule.connection is conn and rule.learning_rule_type is rule_type

    with nengo.Network(seed=seed):
        a, b, e = [nengo.Ensemble(10, 2) for i in range(3)]
        T = np.ones((10, 10))

        r1 = PES()
        c1 = nengo.Connection(a.neurons, b.neurons, learning_rule_type=r1)
        check_rule(c1.learning_rule, c1, r1)

        r2 = [PES(), BCM()]
        c2 = nengo.Connection(a.neurons, b.neurons, learning_rule_type=r2,
                              transform=T)
        assert isinstance(c2.learning_rule, list)
        for rule, rule_type in zip(c2.learning_rule, r2):
            check_rule(rule, c2, rule_type)

        r3 = dict(oja=Oja(), bcm=BCM())
        c3 = nengo.Connection(a.neurons, b.neurons, learning_rule_type=r3,
                              transform=T)
        assert isinstance(c3.learning_rule, dict)
        assert set(c3.learning_rule) == set(r3)  # assert same keys
        for key in r3:
            check_rule(c3.learning_rule[key], c3, r3[key])


def test_voja_encoders(Simulator, nl_nodirect, rng, seed):
    """Tests that voja changes active encoders to the input."""
    n = 200
    learned_vector = np.asarray([0.3, -0.4, 0.6])
    learned_vector /= np.linalg.norm(learned_vector)
    n_change = n // 2  # modify first half of the encoders

    # Set the first half to always fire with random encoders, and the
    # remainder to never fire due to their encoder's dot product with the input
    intercepts = np.asarray([-1]*n_change + [0.99]*(n - n_change))
    rand_encoders = UniformHypersphere(surface=True).sample(
        n_change, len(learned_vector), rng=rng)
    encoders = np.append(
        rand_encoders, [-learned_vector] * (n - n_change), axis=0)

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = nl_nodirect()
        u = nengo.Node(output=learned_vector)
        x = nengo.Ensemble(n, dimensions=len(learned_vector),
                           intercepts=intercepts, encoders=encoders,
                           max_rates=nengo.dists.Uniform(300., 400.),
                           radius=2.0)  # to test encoder scaling

        conn = nengo.Connection(
            u, x, synapse=None, learning_rule_type=Voja(learning_rate=1e-1))
        p_enc = nengo.Probe(conn.learning_rule, 'scaled_encoders')
        p_enc_ens = nengo.Probe(x, 'scaled_encoders')

    with Simulator(m) as sim:
        sim.run(1.0)
    t = sim.trange()
    tend = t > 0.5

    # Voja's rule relies on knowing exactly how the encoders were scaled
    # during the build process, because it modifies the scaled_encoders signal
    # proportional to this factor. Therefore, we should check that its
    # assumption actually holds.
    encoder_scale = (sim.data[x].gain / x.radius)[:, np.newaxis]
    assert np.allclose(sim.data[x].encoders,
                       sim.data[x].scaled_encoders / encoder_scale)

    # Check that the last half kept the same encoders throughout the simulation
    assert np.allclose(
        sim.data[p_enc][0, n_change:], sim.data[p_enc][:, n_change:])
    # and that they are also equal to their originally assigned value
    assert np.allclose(
        sim.data[p_enc][0, n_change:] / encoder_scale[n_change:],
        -learned_vector)

    # Check that the first half converged to the input
    assert np.allclose(
        sim.data[p_enc][tend, :n_change] / encoder_scale[:n_change],
        learned_vector, atol=0.01)
    # Check that encoders probed from ensemble equal encoders probed from Voja
    assert np.allclose(sim.data[p_enc], sim.data[p_enc_ens])


def test_voja_modulate(Simulator, nl_nodirect, seed):
    """Tests that voja's rule can be modulated on/off."""
    n = 200
    learned_vector = np.asarray([0.5])

    def control_signal(t):
        """Modulates the learning on/off."""
        return 0 if t < 0.5 else -1

    m = nengo.Network(seed=seed)
    with m:
        m.config[nengo.Ensemble].neuron_type = nl_nodirect()
        control = nengo.Node(output=control_signal)
        u = nengo.Node(output=learned_vector)
        x = nengo.Ensemble(n, dimensions=len(learned_vector))

        conn = nengo.Connection(
            u, x, synapse=None, learning_rule_type=Voja(post_synapse=None))
        nengo.Connection(control, conn.learning_rule, synapse=None)

        p_enc = nengo.Probe(conn.learning_rule, 'scaled_encoders')

    with Simulator(m) as sim:
        sim.run(1.0)
    tend = sim.trange() > 0.5

    # Check that encoders stop changing after 0.5s
    assert np.allclose(sim.data[p_enc][tend], sim.data[p_enc][-1])

    # Check that encoders changed during first 0.5s
    i = np.where(tend)[0][0]  # first time point after changeover
    assert not np.allclose(sim.data[p_enc][0], sim.data[p_enc][i])


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = PES(learning_rate=2e-3, pre_synapse=4e-3)
    b = PES(learning_rate=2e-3, pre_synapse=4e-3)
    c = PES(learning_rate=2e-3, pre_synapse=5e-3)

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely

    with pytest.raises((ValueError, RuntimeError)):
        a.learning_rate = 1e-1


def test_pes_direct_errors():
    """Test that applying a learning rule to a direct ensemble errors."""
    with nengo.Network():
        pre = nengo.Ensemble(10, 1, neuron_type=nengo.Direct())
        post = nengo.Ensemble(10, 1)
        conn = nengo.Connection(pre, post)
        with pytest.raises(ValidationError):
            conn.learning_rule_type = nengo.PES()


def test_custom_type(Simulator):
    """Test with custom learning rule type.

    A custom learning type may have ``size_in`` not equal to 0, 1, or None.
    """

    class TestRule(nengo.learning_rules.LearningRuleType):
        modifies = 'decoders'

        def __init__(self):
            super(TestRule, self).__init__(1.0, size_in=3)

    @Builder.register(TestRule)
    def build_test_rule(model, test_rule, rule):
        error = Signal(np.zeros(rule.connection.size_in))
        model.add_op(Reset(error))
        model.sig[rule]['in'] = error[:rule.size_in]

        model.add_op(Copy(error, model.sig[rule]['delta']))

    with nengo.Network() as net:
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(10, 1)
        conn = nengo.Connection(a.neurons, b, transform=np.zeros((1, 10)),
                                learning_rule_type=TestRule())

        err = nengo.Node([1, 2, 3])
        nengo.Connection(err, conn.learning_rule, synapse=None)

        p = nengo.Probe(conn, 'weights')

    with Simulator(net) as sim:
        sim.run(sim.dt * 5)

    assert np.allclose(sim.data[p][:, 0, :3],
                       np.outer(np.arange(1, 6), np.arange(1, 4)))
    assert np.allclose(sim.data[p][:, :, 3:], 0)


@pytest.mark.parametrize(
    "LearningRule", (nengo.PES, nengo.BCM, nengo.Voja, nengo.Oja))
def test_tau_deprecation(LearningRule):
    params = [("pre_tau", "pre_synapse"),
              ("post_tau", "post_synapse"),
              ("theta_tau", "theta_synapse")]
    kwargs = {}
    for i, (p0, p1) in enumerate(params):
        if hasattr(LearningRule, p0):
            kwargs[p0] = i

    with pytest.warns(DeprecationWarning):
        l_rule = LearningRule(learning_rate=0, **kwargs)

    for i, (p0, p1) in enumerate(params):
        if hasattr(LearningRule, p0):
            assert getattr(l_rule, p0) == i
            assert getattr(l_rule, p1) == Lowpass(i)


def test_slicing(Simulator, seed):
    with nengo.Network(seed=seed) as model:
        a = nengo.Ensemble(30, 1)
        b = nengo.Ensemble(30, 2)
        conn = nengo.Connection(
            a, b, learning_rule_type=PES(), function=lambda x: (0, 0))
        nengo.Connection(nengo.Node(1.), a)

        err1 = nengo.Node(lambda t, x: x - 0.75, size_in=1)
        nengo.Connection(b[0], err1)
        nengo.Connection(err1, conn.learning_rule[0])

        err2 = nengo.Node(lambda t, x: x + 0.5, size_in=1)
        nengo.Connection(b[1], err2)
        nengo.Connection(err2, conn.learning_rule[1])

        p = nengo.Probe(b, synapse=0.03)

    with nengo.Simulator(model) as sim:
        sim.run(1.)

    t = sim.trange() > 0.8
    assert np.allclose(sim.data[p][t, 0], 0.75, atol=0.15)
    assert np.allclose(sim.data[p][t, 1], -0.5, atol=0.15)



import pytest

import nengo
from nengo.utils.probe import probe_all


@pytest.mark.parametrize("recursive", [False, True])
def test_probe_all_recursive(recursive):

    model = nengo.Network(label='test_probing')
    with model:
        ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
        node1 = nengo.Node(output=[0])
        conn = nengo.Connection(node1, ens1)
        subnet = nengo.Network(label='subnet')

        with subnet:
            ens2 = nengo.Ensemble(n_neurons=1, dimensions=1)
            node2 = nengo.Node(output=[0])

    probes = probe_all(model, recursive=recursive)

    # test top level probing
    total_number_probes1 = len(
        ens1.probeable) + len(node1.probeable) + len(conn.probeable)
    assert(len(model.probes) == total_number_probes1)

    # test dictionary
    assert(len(probes[ens1]) == len(ens1.probeable))
    assert(len(probes[node1]) == len(node1.probeable))
    assert(len(probes[conn]) == len(conn.probeable))

    # test recursive probing
    if recursive:
        total_number_probes2 = len(ens2.probeable) + len(node2.probeable)
        assert(len(subnet.probes) == total_number_probes2)
        assert(len(probes[ens2]) == len(ens2.probeable))
        assert(len(probes[node2]) == len(node2.probeable))
    else:
        assert(len(subnet.probes) == 0)
        assert ens2 not in probes


def test_probe_all_options():
    model = nengo.Network(label='test_probing')
    with model:
        ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
        node1 = nengo.Node(output=[0])
        nengo.Connection(node1, ens1)
        subnet = nengo.Network(label='subnet')

        with subnet:
            nengo.Ensemble(n_neurons=1, dimensions=1)
            nengo.Node(output=[0])

    probe_all(model, recursive=True, probe_options={
        nengo.Ensemble: ['decoded_output']})

    # only probes spikes and decoded output of the ensembles
    assert(len(model.probes) == 1)
    assert(len(subnet.probes) == 1)


def test_probe_all_kwargs():
    model = nengo.Network(label='test_probing')
    with model:
        ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
        node1 = nengo.Node(output=[0])
        nengo.Connection(node1, ens1)
        subnet = nengo.Network(label='subnet')

        with subnet:
            nengo.Ensemble(n_neurons=1, dimensions=1)
            nengo.Node(output=[0])

    probe_all(model, recursive=True, sample_every=0.1, seed=10)
    for probe in model.probes + subnet.probes:
        assert probe.sample_every == 0.1
        assert probe.seed == 10



import numpy as np

import nengo
from nengo.spa import Vocabulary, Input
from nengo.spa.utils import similarity
from nengo.spa.assoc_mem import AssociativeMemory


def test_am_spa_interaction(Simulator, seed, rng):
    """Make sure associative memory interacts with other SPA modules."""
    D = 16
    vocab = Vocabulary(D, rng=rng)
    vocab.parse('A+B+C+D')

    D2 = int(D / 2)
    vocab2 = Vocabulary(D2, rng=rng)
    vocab2.parse('A+B+C+D')

    def input_func(t):
        return '0.49*A' if t < 0.5 else '0.79*A'

    with nengo.spa.SPA(seed=seed) as m:
        m.buf = nengo.spa.Buffer(D)
        m.input = nengo.spa.Input(buf=input_func)

        m.am = AssociativeMemory(vocab, vocab2,
                                 input_keys=['A', 'B', 'C'],
                                 output_keys=['B', 'C', 'D'],
                                 default_output_key='A',
                                 threshold=0.5,
                                 inhibitable=True,
                                 wta_output=True,
                                 threshold_output=True)

        cortical_actions = nengo.spa.Actions('am = buf')
        m.c_act = nengo.spa.Cortical(cortical_actions)

    # Check to see if model builds properly. No functionality test needed
    with Simulator(m):
        pass


def test_am_spa_keys_as_expressions(Simulator, plt, seed, rng):
    """Provide semantic pointer expressions as input and output keys."""
    D = 64

    vocab_in = Vocabulary(D, rng=rng)
    vocab_out = Vocabulary(D, rng=rng)

    vocab_in.parse('A+B')
    vocab_out.parse('C+D')

    in_keys = ['A', 'A*B']
    out_keys = ['C*D', 'C+D']

    with nengo.spa.SPA(seed=seed) as model:
        model.am = AssociativeMemory(input_vocab=vocab_in,
                                     output_vocab=vocab_out,
                                     input_keys=in_keys,
                                     output_keys=out_keys)

        model.inp = Input(am=lambda t: 'A' if t < 0.1 else 'A*B')

        in_p = nengo.Probe(model.am.input)
        out_p = nengo.Probe(model.am.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.2)

    # Specify t ranges
    t = sim.trange()
    t_item1 = (t > 0.075) & (t < 0.1)
    t_item2 = (t > 0.175) & (t < 0.2)

    # Modify vocabularies (for plotting purposes)
    vocab_in.add(in_keys[1], vocab_in.parse(in_keys[1]).v)
    vocab_out.add(out_keys[0], vocab_out.parse(out_keys[0]).v)

    plt.subplot(2, 1, 1)
    plt.plot(t, similarity(sim.data[in_p], vocab_in))
    plt.ylabel("Input: " + ', '.join(in_keys))
    plt.legend(vocab_in.keys, loc='best')
    plt.ylim(top=1.1)
    plt.subplot(2, 1, 2)
    plt.plot(t, similarity(sim.data[out_p], vocab_out))
    plt.plot(t[t_item1], np.ones(t.shape)[t_item1] * 0.9, c='r', lw=2)
    plt.plot(t[t_item2], np.ones(t.shape)[t_item2] * 0.91, c='g', lw=2)
    plt.plot(t[t_item2], np.ones(t.shape)[t_item2] * 0.89, c='b', lw=2)
    plt.ylabel("Output: " + ', '.join(out_keys))
    plt.legend(vocab_out.keys, loc='best')

    assert np.mean(similarity(sim.data[out_p][t_item1],
                              vocab_out.parse(out_keys[0]).v,
                              normalize=True)) > 0.9
    assert np.mean(similarity(sim.data[out_p][t_item2],
                              vocab_out.parse(out_keys[1]).v,
                              normalize=True)) > 0.9




# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '1.3.0'
version = '1.3.0'
full_version = '1.3.0'
git_revision = 'e1e44d12637997606b1bcc0c6de232349e11eee0'
release = True

if not release:
    version = full_version



import numpy as np
import pytest

from scipy.linalg import inv

import nengo

from nengolib.signal.realizers import (
    _realize, Identity, Balanced, Hankel, L1Norm, H2Norm)
from nengolib import Lowpass, Alpha, Network
from nengolib.networks import LinearNetwork
from nengolib.signal import (
    ss_equal, sys_equal, state_norm, balanced_transformation)
from nengolib.synapses import Bandpass, Highpass, PadeDelay


@pytest.mark.parametrize("radii", [1, 3, [1.5, 0.2]])
def test_func_realize(radii):
    sys = Alpha(0.1)

    T = np.asarray([[1., 2.], [0, -1.]])

    for Tinv in (None, inv(T)):
        realize_result = _realize(sys, radii, T, Tinv)

        assert realize_result.sys is sys
        assert np.allclose(inv(realize_result.T), realize_result.Tinv)

        rsys = realize_result.realization
        assert ss_equal(rsys, sys.transform(realize_result.T))

        # Check that the state vector are related by T
        length = 1000
        dt = 0.001
        x_old = np.asarray([sub.impulse(length, dt) for sub in sys])
        x_new = np.asarray([sub.impulse(length, dt) for sub in rsys])

        r = np.atleast_2d(np.asarray(radii).T).T
        assert np.allclose(np.dot(T, x_new * r), x_old)


def test_invalid_realize():
    sys = Lowpass(0.1)

    with pytest.raises(ValueError):
        _realize(sys, radii=[[1]], T=np.eye(len(sys)))

    with pytest.raises(ValueError):
        _realize(sys, radii=[1, 2], T=np.eye(len(sys)))


@pytest.mark.parametrize("radii", [0.5, 1, [1.5, 0.2]])
def test_identity(radii):
    sys = Alpha(0.1)

    identity = Identity()
    assert repr(identity) == "Identity()"

    I = np.eye(len(sys))
    realize_result = identity(sys, radii)
    assert realize_result.sys is sys
    assert np.allclose(realize_result.T, I * radii)
    assert np.allclose(realize_result.Tinv, inv(I * radii))

    rsys = realize_result.realization
    assert ss_equal(rsys, sys.transform(realize_result.T))

    # Check that it's still the same system, even though different matrices
    assert sys_equal(sys, rsys)
    if radii == 1:
        assert ss_equal(sys, rsys)
    else:
        assert not np.allclose(sys.B, rsys.B)
        assert not np.allclose(sys.C, rsys.C)

    # Check that the state vectors have scaled power
    assert np.allclose(state_norm(sys) / radii, state_norm(rsys))


@pytest.mark.parametrize("sys", [PadeDelay(0.1, 4), PadeDelay(0.05, 5, 5)])
def test_balreal_normalization(sys):
    radii = np.arange(len(sys)) + 1

    balanced = Balanced()
    assert repr(balanced) == "Balanced()"

    realizer_result = balanced(sys, radii)

    T, Tinv, _ = balanced_transformation(sys)

    assert np.allclose(realizer_result.T / radii[None, :], T)
    assert np.allclose(realizer_result.Tinv * radii[:, None], Tinv)
    assert np.allclose(inv(T), Tinv)

    assert sys_equal(sys, realizer_result.realization)


def _test_normalization(Simulator, sys, rng, realizer, l1_lower,
                        lower, radius=5.0, dt=0.0001, T=1.0, eps=1e-5):
    response = sys.X.impulse(int(T / dt), dt=dt)
    assert np.allclose(response[-10:], 0)
    l1_norms = radius * np.sum(abs(response * dt), axis=0)

    with Network() as model:
        stim = nengo.Node(output=lambda t: rng.choice([-radius, radius])
                          if t < T/2 else radius)
        tau = 0.02
        subnet = LinearNetwork(sys, n_neurons_per_ensemble=1, synapse=tau,
                               dt=dt, input_synapse=tau,
                               radii=radius, realizer=realizer,
                               neuron_type=nengo.neurons.Direct())
        nengo.Connection(stim, subnet.input, synapse=None)
        p = nengo.Probe(subnet.state.output, synapse=None)

        assert np.allclose(inv(subnet.realizer_result.T),
                           subnet.realizer_result.Tinv)

    trans = subnet.realizer_result.T
    assert trans.shape == (len(sys), len(sys))
    est_worst_x = np.diagonal(trans)
    assert np.allclose(np.diag(est_worst_x), trans)  # make sure diagonal
    assert est_worst_x.shape == (len(sys),)

    assert ((l1_lower*est_worst_x <= l1_norms) |
            (est_worst_x <= eps)).all()
    assert (l1_norms <= est_worst_x + eps).all()

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    # lower bound includes both approximation error and the gap between
    # random {-1, 1} flip-flop inputs and the true worst-case input
    worst_x = np.max(abs(sim.data[p]), axis=0)
    assert (lower <= worst_x + eps).all()
    assert (worst_x <= 1 + eps).all()


def test_l1_repr():
    assert (repr(L1Norm(rtol=.1, max_length=10)) ==
            "L1Norm(rtol=0.1, max_length=10)")


@pytest.mark.parametrize("sys,lower", [
    (Lowpass(0.005), 1.0), (Alpha(0.01), 0.3), (Bandpass(50, 5), 0.05),
    (Highpass(0.01, 4), 0.1), (PadeDelay(0.1, 2, 2), 0.3)])
def test_hankel_normalization(Simulator, rng, sys, lower):
    _test_normalization(Simulator, sys, rng, Hankel(),
                        l1_lower=0.5, lower=lower)


@pytest.mark.parametrize("radius", [0.5, 5, 10])
@pytest.mark.parametrize("sys", [Lowpass(0.005)])
def test_normalization_radius(Simulator, rng, sys, radius):
    _test_normalization(Simulator, sys, rng, L1Norm(), radius=radius,
                        l1_lower=1-1e-3, lower=1.0)


@pytest.mark.parametrize("sys,lower", [
    (Alpha(0.01), 0.4), (Bandpass(50, 5), 0.1), (Highpass(0.01, 4), 0.2),
    (PadeDelay(0.1, 2, 2), 0.4), (PadeDelay(0.1, 3), 0.3),
    (PadeDelay(0.1, 4, 4), 0.15)])
def test_l1_normalization_crossing(Simulator, rng, sys, lower):
    # note the lower bounds are higher than those for the hankel norm tests
    # for the same systems
    # TODO: highpass performs unexpectedly poorly (0.9 is too tolerant)
    _test_normalization(Simulator, sys, rng, L1Norm(),
                        l1_lower=0.9, lower=lower)


@pytest.mark.parametrize("sys", [
    Alpha(0.01), Bandpass(50, 5), Highpass(0.01, 4),
    PadeDelay(0.1, 2, 2), PadeDelay(1.0, 10),
    PadeDelay(2.0, 4, 4)])
def test_h2_realizer(sys):
    r1 = Hankel()(sys)
    r2 = H2Norm()(sys)

    assert np.allclose(np.diag(np.diagonal(r2.T)), r2.T)  # make sure diagonal
    s1 = np.diag(r1.T)
    s2 = np.diag(r2.T)

    # Hankel and H2Norm have the same shape and order of magnitude
    assert (0.3*s1 <= s2).all()
    assert (6*s1 >= s2).all()
    assert (np.sign(np.diff(s1)) == np.sign(np.diff(s1))).all()
